package ch.psi.daq.rest.config;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;

import com.fasterxml.jackson.annotation.JsonInclude.Include;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.psi.daq.common.statistic.StorelessStatistics;
import ch.psi.daq.domain.cassandra.ChannelEvent;
import ch.psi.daq.hazelcast.config.HazelcastClientConfig;
import ch.psi.daq.hazelcast.config.HazelcastConfig;
import ch.psi.daq.rest.model.PropertyFilterMixin;
import ch.psi.daq.rest.response.ResponseStreamWriter;

@Configuration
@PropertySource(value = {"classpath:rest.properties"})
@PropertySource(value = {"file:${user.home}/.config/daq/rest.properties"}, ignoreResourceNotFound = true)
public class RestConfig {

   @Autowired
   private Environment env;

   // a nested configuration
   // this guarantees that the ordering of the properties file is as expected
   // see:
   // https://jira.spring.io/browse/SPR-10409?focusedCommentId=101393&page=com.atlassian.jira.plugin.system.issuetabpanels:comment-tabpanel#comment-101393
   @Configuration
   @Import({HazelcastConfig.class, HazelcastClientConfig.class})
   static class InnerConfiguration {
   }

   @Bean
   public JsonFactory jsonFactory() {
      return new JsonFactory();
   }

   @Bean
   public ResponseStreamWriter responseStreamWriter() {
      return new ResponseStreamWriter();
   }

   @Bean
   public ObjectMapper objectMapper() {
      ObjectMapper mapper = new ObjectMapper();
      // only include non-null values
      mapper.setSerializationInclusion(Include.NON_NULL);
      // Mixin which is used dynamically to filter out which properties get serialised and which
      // won't. This way, the user can specify which columns are to be received.
      mapper.addMixIn(ChannelEvent.class, PropertyFilterMixin.class);
      mapper.addMixIn(StorelessStatistics.class, PropertyFilterMixin.class);
      return mapper;
   }

}
