package ch.psi.daq.rest.controller;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.function.LongFunction;
import java.util.function.LongUnaryOperator;
import java.util.stream.Collectors;
import java.util.stream.LongStream;
import java.util.stream.Stream;

import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.util.Assert;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import ch.psi.daq.cassandra.writer.CassandraWriter;
import ch.psi.daq.domain.DataType;
import ch.psi.daq.domain.cassandra.ChannelEvent;
import ch.psi.daq.domain.cassandra.DataEvent;
import ch.psi.daq.hazelcast.query.AbstractQuery;
import ch.psi.daq.hazelcast.query.PulseRangeQuery;
import ch.psi.daq.hazelcast.query.TimeRangeQuery;
import ch.psi.daq.hazelcast.query.processor.QueryProcessor;
import ch.psi.daq.rest.response.ResponseStreamWriter;

@RestController
public class DaqRestController {

   private static final Logger logger = LoggerFactory.getLogger(DaqRestController.class);
   
   @Autowired
   private CassandraWriter cassandraWriter;
   
   @Autowired
   private ResponseStreamWriter responseStreamWriter;

   @Autowired
   private QueryProcessor queryProcessor;

   
   @RequestMapping(value = "/pulserange")
   public void pulseRange(@RequestBody PulseRangeQuery query, HttpServletResponse res) throws IOException {

      logger.debug("PulseRangeQuery received: {}", query);

      executeQuery(query, res);
   }

   
   @RequestMapping(value = "/timerange")
   public void timeRange(@RequestBody TimeRangeQuery query, HttpServletResponse res) throws IOException {
      
      logger.debug("TimeRangeQuery received: {}", query);
      
      executeQuery(query, res);
   }

   
   private void executeQuery(AbstractQuery query, HttpServletResponse res) throws IOException {

      // all the magic happens here
      Map<String, Stream<? extends DataEvent>> process = queryProcessor.process(query);

      Stream<DataEvent> flatStreams = process.values().stream().flatMap(s -> {
         return s;
      });

      // write the response back to the client using java 8 streams
      responseStreamWriter.respond(flatStreams, query, res);
   }

   // ==========================================================================================
   // TODO this is simply a method for initial / rudimentary testing - remove once further evolved
   @RequestMapping(value = "/write")
   public long writeDummyEntry() {
      
      long startIndex = System.currentTimeMillis();
      
      writeData(3,startIndex, 100, new String[]{"channel1", "channel2"});
      
      return startIndex;
   }
   
   private void writeData(int dataReplication, long startIndex, long nrOfElements,
         String... channelNames) {
      writeData(dataReplication, startIndex, nrOfElements,
            i -> 2 * i,
            i -> 2 * i,
            i -> 2 * i,
            i -> 2 * i,
            i -> 2 * i,
            i -> Long.valueOf(2 * i),
            channelNames);
   }
   
   private <T> void writeData(int dataReplication, long startIndex, long nrOfElements,
         LongUnaryOperator iocMillis, LongUnaryOperator iocNanos, LongUnaryOperator pulseIds,
         LongUnaryOperator globalMillis, LongUnaryOperator globalNanos, LongFunction<T> valueFunction,
         String... channelNames) {
      
      Assert.notNull(channelNames);

      CompletableFuture<Void> future;

      List<ChannelEvent> events =
            Arrays.stream(channelNames)
                  .parallel()
                  .flatMap(
                        channelName -> {
                           Stream<ChannelEvent> stream =
                                 LongStream
                                       .range(startIndex, startIndex + nrOfElements)
                                       .parallel()
                                       .mapToObj(
                                             i -> {
                                                Object value = valueFunction.apply(i);
                                                return new ChannelEvent(channelName,
                                                      iocMillis.applyAsLong(i),
                                                      iocNanos.applyAsLong(i),
                                                      pulseIds.applyAsLong(i),
                                                      globalMillis.applyAsLong(i),
                                                      globalNanos.applyAsLong(i),
                                                      value,
                                                      DataType.getTypeName(value.getClass())
                                                );
                                             }
                                       );

                           return stream;
                        })
                  .collect(Collectors.toList());

      future = cassandraWriter.writeAsync(dataReplication, (int) TimeUnit.HOURS.toSeconds(1), events);
      future.join();
   }
}
